/* vim: set sw=8: -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * fn-japanese.c:  japanese function
 *
 * Author:
 *   Yukihiro Nakai (ynakai@redhat.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gnumeric-config.h>
#include <gnumeric-i18n.h>
#include <gnumeric.h>
#include <func.h>

#include <parse-util.h>
#include <cell.h>
#include <sheet.h>
#include <format.h>
#include <workbook.h>
#include <expr.h>
#include <value.h>

#include "plugin.h"
#include "plugin-util.h"
#include "module-plugin-defs.h"

#include "jisasc.h"

GNUMERIC_MODULE_PLUGIN_INFO_DECL;

/***************************************************************************/

static char const *help_jis = {
	N_("@FUNCTION=JIS\n"
	   "@SYNTAX=JIS(text)\n"

	   "@DESCRIPTION="
	   "JIS converts half-width chars in @text to full-width char "
	   "and return.\n"
	   "\n"
	   "* This function is Excel (Japanese version) "
           "compatible.\n"
	   "\n"
	   "@EXAMPLES=\n"
	   "\n"
	   "@SEEALSO=")
};

static Value *
gnumeric_jis (FunctionEvalInfo *ei, Value **argv)
{
	gchar *os;
	gchar* zen_utf8;

	os = g_strdup(value_peek_string(argv[0]));
	zen_utf8 = han2zen(os);
	g_free(os);
	return value_new_string_nocopy(zen_utf8);

}

static char const *help_asc = {
	N_("@FUNCTION=ASC\n"
	   "@SYNTAX=ASC(text)\n"

	   "@DESCRIPTION="
	   "ASC converts full-width chars in @text to half-width char "
	   "and return.\n"
	   "\n"
	   "* This function is Excel (Japanese version) "
           "compatible.\n"
	   "\n"
	   "@EXAMPLES=\n"
	   "\n"
	   "@SEEALSO=")
};

static Value *
gnumeric_asc (FunctionEvalInfo *ei, Value **argv)
{
	gchar *os;
	Value* result;
	gchar* han_utf8;

	os = g_strdup(value_peek_string(argv[0]));
	han_utf8 = zen2han(os);
	result = value_new_string(han_utf8);

	g_free(os);
	g_free(han_utf8);

	return result;
}


/***************************************************************************/

const GnmFuncDescriptor japanese_functions[] = {
	{ "jis", "s", N_("text"), &help_jis,
	  gnumeric_jis, NULL, NULL, NULL, NULL,
	  GNM_FUNC_SIMPLE,
	  GNM_FUNC_IMPL_STATUS_COMPLETE, GNM_FUNC_TEST_STATUS_BASIC },
	{ "asc", "s", N_("text"), &help_asc,
	  gnumeric_asc, NULL, NULL, NULL, NULL,
	  GNM_FUNC_SIMPLE,
	  GNM_FUNC_IMPL_STATUS_COMPLETE, GNM_FUNC_TEST_STATUS_BASIC },
        {NULL}
};

void
plugin_init (void)
{
	convtable_init();
}
