#include <glib.h>
#include <stdio.h>
#define __USE_GNU /* To define strndup.. ??? */
#include <string.h>
#include "jisasc.h"

#include "convtable.h"

GHashTable* zen2han_ht = NULL;
GHashTable* han2zen_ht = NULL;

#define DAKUTEN_UNICHAR    0xFF9E /* Dakuten */
#define HANDAKUTEN_UNICHAR 0xFF9F /* Han-Dakuten */

#define g_utf8_get_2nd_char(p) \
  (g_utf8_get_char(g_utf8_offset_to_pointer(p,1)))

void
convtable_init() {
  int i;

  zen2han_ht = g_hash_table_new(g_str_hash, g_str_equal);
  han2zen_ht = g_hash_table_new(g_str_hash, g_str_equal);

  /* Add Data */
  for(i=0;i<sizeof(twoway_table)/sizeof(twoway_table[0]);i++) {
    g_hash_table_insert(zen2han_ht,
                        (gpointer)twoway_table[i].lstr,
                        (gpointer)twoway_table[i].rstr);
    g_hash_table_insert(han2zen_ht,
                        (gpointer)twoway_table[i].rstr,
                        (gpointer)twoway_table[i].lstr);
  }
  for(i=0;i<sizeof(zen2han_table)/sizeof(zen2han_table[0]);i++) {
    g_hash_table_insert(zen2han_ht,
                        (gpointer)zen2han_table[i].lstr,
                        (gpointer)zen2han_table[i].rstr);
  }
  for(i=0;i<sizeof(han2zen_table)/sizeof(han2zen_table[0]);i++) {
    g_hash_table_insert(han2zen_ht,
                        (gpointer)han2zen_table[i].rstr,
                        (gpointer)han2zen_table[i].lstr);
  }
}

static gint
compare_simple(gconstpointer a, gconstpointer b) {
  return (gint)(a - b);
}

static gchar*
g_utf8_pop_char(const gchar *p) {
  g_return_val_if_fail(p != NULL, NULL);

  if( g_utf8_strlen(p, -1) <= 1 )
    return (gchar*)strdup(p);

  return (gchar*)strndup(p, strlen(p) - strlen(g_utf8_offset_to_pointer(p, 1)));
}

static gchar*
g_utf8_pop_2chars(const gchar *p) {
  g_return_val_if_fail(p != NULL, NULL);

  if( g_utf8_strlen(p, -1) <= 1 )
    return (gchar*)strdup(p);

  return (gchar*)strndup(p, strlen(p) - strlen(g_utf8_offset_to_pointer(p, 2)));
}

static gchar*
g_str_realloc0(gchar* str, gulong len) {
  gulong current_size = strlen(str);
  gchar* result = (gchar*)g_realloc(str, len);

  if( current_size > len )
    return result;

  memset(result+current_size, 0, len - current_size);

  return result;
}

gchar*
zen2han(gchar* utf8str) {
  gchar* p = utf8str;
  gchar* result;
  size_t buflen;

  if( utf8str == NULL || *utf8str == '\0' )
    return g_strdup("");

  if( !g_utf8_validate(utf8str, -1, NULL) )
	  return g_strdup(utf8str);

  buflen = strlen(utf8str);
  result = g_new0(gchar, buflen+1);

  while( p != NULL && *p != '\0' ) {
    gchar* first = g_utf8_pop_char(p);
    gchar* first_han = (gchar*)g_hash_table_lookup(zen2han_ht, first);
    if( first_han != NULL ) {
      if( strlen(result)+strlen(first_han) > buflen ) {
        buflen = strlen(result)+strlen(first_han);
        result = g_str_realloc0(result, buflen+1);
      }
      strcat(result, first_han);
    } else {
      if( strlen(result)+strlen(first) > buflen ) {
        buflen = strlen(result)+strlen(first);
        result = g_str_realloc0(result, buflen+1);
      }
      strcat(result, first);
    }
    g_free(first);
    p = g_utf8_offset_to_pointer(p, 1);
  }

  return result;
}

gchar*
han2zen(gchar* utf8str) {
  gchar* p = utf8str;
  gchar* result;
  size_t buflen;

  if( utf8str == NULL || *utf8str == '\0' )
    return g_strdup("");

  if( !g_utf8_validate(utf8str, -1, NULL) )
	  return g_strdup(utf8str);

  buflen = strlen(utf8str);
  result = g_new0(gchar, buflen+1);

  while( p != NULL && *p != '\0' ) {
    gchar* first = NULL;
    gchar* first_zen;
    gboolean second_is_dakuten = FALSE;

    if( g_utf8_strlen(p, -1) >= 2 ) {
      first = g_utf8_pop_2chars(p);
      first_zen = (gchar*)g_hash_table_lookup(han2zen_ht, first);
      second_is_dakuten = (first_zen != NULL);
    }

    if( !second_is_dakuten ) {
      first = g_utf8_pop_char(p);
      first_zen = (gchar*)g_hash_table_lookup(han2zen_ht, first);
    }

    if( first_zen != NULL ) {
      if( strlen(first_zen)+strlen(result) > buflen ) {
	buflen = strlen(first_zen)+strlen(result);
	result = g_str_realloc0(result, buflen+1);
      }
      strcat(result, first_zen);
    } else {
      if( strlen(first)+strlen(result) > buflen ) {
	buflen = strlen(first)+strlen(result);
	result = g_str_realloc0(result, buflen+1);
      }
      strcat(result, first);
    }
    g_free(first);

    p = g_utf8_offset_to_pointer(p, second_is_dakuten ? 2 : 1);
  }
  
  return result;
}

void
dump_2way_text() {
  int i;
  gunichar unic;
  gchar utf8buf[7];
  GList* key_list = NULL;

  for(i=0;i<sizeof(twoway_table)/sizeof(twoway_table[0]);i++) {
    unic = g_utf8_get_char(twoway_table[i].lstr);
    key_list = g_list_insert_sorted(key_list, (gpointer)unic, compare_simple);
  }

  for(i=0;i<g_list_length(key_list);i++) {
    unic = (gunichar)g_list_nth_data(key_list, i);
    memset(utf8buf, 0, 7);
    g_unichar_to_utf8(unic, utf8buf);
    printf("%s\t%s\n", utf8buf, (gchar*)g_hash_table_lookup(zen2han_ht, utf8buf));
  }
}

void
dump_header() {
  int i;
  gunichar unic;
  gchar utf8buf[7];
  GList* key_list = NULL;
  gchar* hankaku;

  for(i=0;i<sizeof(twoway_table)/sizeof(twoway_table[0]);i++) {
    unic = g_utf8_get_char(twoway_table[i].lstr);
    key_list = g_list_insert_sorted(key_list, (gpointer)unic, compare_simple);
  }

  g_print("#ifndef _CONVTABLE_H\n"
          "#define _CONVTABLE_H\n"
          "\n"
          "#include <glib.h>\n"
          "\n"
          "typedef struct _strpair {\n"
          "  const gchar* lstr;\n"
          "  const gchar* rstr;\n"
          "} strpair;\n"
          "\n"
          "strpair twoway_table[] = {\n");

  for(i=0;i<g_list_length(key_list);i++) {
    unic = (gunichar)g_list_nth_data(key_list, i);
    memset(utf8buf, 0, 7);
    g_unichar_to_utf8(unic, utf8buf);
    hankaku = g_hash_table_lookup(zen2han_ht, utf8buf);
    printf("  {\"%s\", ", utf8buf);
    if( ((guchar*)hankaku)[0] == '\"' && strlen(hankaku) == 1 )
      printf("\"\\%s\" }, ", hankaku);
    else if( ((guchar*)hankaku)[0] == '\\' && strlen(hankaku) == 1 )
      printf("\"\\%s\" }, ", hankaku);
    else
      printf("\"%s\" }, ", hankaku);
    printf("\t/* 0x%04X */\n", unic);
  }
  g_print("};\n\n");

  g_list_free(key_list);
  key_list = NULL;
  for(i=0;i<sizeof(zen2han_table)/sizeof(zen2han_table[0]);i++) {
    unic = g_utf8_get_char(zen2han_table[i].lstr);
    key_list = g_list_insert_sorted(key_list, (gpointer)unic, compare_simple);
  }

  g_print("strpair zen2han_table[] = {\n");
  for(i=0;i<g_list_length(key_list);i++) {
    unic = (gunichar)g_list_nth_data(key_list, i);
    memset(utf8buf, 0, 7);
    g_unichar_to_utf8(unic, utf8buf);
    hankaku = g_hash_table_lookup(zen2han_ht, utf8buf);
    printf("  {\"%s\", ", utf8buf);
    if( ((guchar*)hankaku)[0] == '\"' && strlen(hankaku) == 1 )
      printf("\"\\%s\" }, ", hankaku);
    else if( ((guchar*)hankaku)[0] == '\\' && strlen(hankaku) == 1 )
      printf("\"\\%s\" }, ", hankaku);
    else
      printf("\"%s\" }, ", hankaku);
    printf("\t/* 0x%04X */\n", unic);
  }
  g_print("};\n\n");

  g_list_free(key_list);
  key_list = NULL;
  for(i=0;i<sizeof(han2zen_table)/sizeof(han2zen_table[0]);i++) {
    gchar* utf8line;
    unic = g_utf8_get_char(han2zen_table[i].lstr);
    if( han2zen_table[i].rstr[0] == '\\' || han2zen_table[i].rstr[0] == '\"' )
      utf8line = g_strdup_printf("  { \"%s\", \"\\%s\" },\t/* 0x%04X */\n", han2zen_table[i].lstr, han2zen_table[i].rstr, unic);
    else
      utf8line = g_strdup_printf("  { \"%s\", \"%s\" },\t/* 0x%04X */\n", han2zen_table[i].lstr, han2zen_table[i].rstr, unic);
    key_list = g_list_insert_sorted(key_list, utf8line, (GCompareFunc)g_utf8_collate);
  }

  g_print("strpair han2zen_table[] = {\n");
  for(i=0;i<g_list_length(key_list);i++) {
    gchar* utf8line = g_list_nth_data(key_list, i);
    printf("%s", utf8line);
    g_free(utf8line);
  }
  g_print("};\n\n");

  g_print("#endif /* _CONVTABLE_H */\n");

  g_list_free(key_list);
}
